/*
 * surface.cc -- Shaded surface plotting
 *
 * This file is part of ePiX, a preprocessor for creating high-quality
 * line figures in LaTeX
 *
 * Version 1.0.7
 * Last Change: May 19, 2006
 */

/*
 * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
 * Andrew D. Hwang <rot 13 nujnat at zngupf dot ubylpebff dot rqh>
 * Department of Mathematics and Computer Science
 * College of the Holy Cross
 * Worcester, MA, 01610-2395, USA
 */

/*
 * ePiX is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ePiX is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ePiX; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include "globals.h"
#include "camera.h"
#include "triples.h"
#include "domain.h"
#include "curves.h"
#include "output.h"
#include "surface.h"

namespace ePiX {
  facet::facet(P p1, P p2, P p3, P p4) : pt1(p1), pt2(p2), pt3(p3), pt4(p4)
  {
    center=0.25*(pt1 + pt2 + pt3 + pt4);
    direction=center-camera.get_viewpt();
    distance=norm(direction);
  }

  bool facet::front_facing(void) const
  {
    return ((camera.get_viewpt()-center)|(pt3-center)) > -EPIX_EPSILON;
  }

  void facet::draw(void) const
  {
    bool my_fill=epix::fill_paths; // get fill state
    if (my_fill) // calculate cosine^2 of normal angle
      {
	P normal = (pt2 - pt1)*(pt4 - pt1);
	normal *= recip(norm(normal));

	gray(0.75*(1-pow(normal|(recip(distance)*direction), 2)));
      }

    fill();
    ePiX::quad(pt1, pt2, pt3, pt4); 
    fill(my_fill); // restore state
  }


  void surface(P F(double, double), const domain& R)
  {
    if (R.dx3() > 0)
      epix_warning("surface() ignores domain's third coordinate");

    int i_max = (R.dx1() > 0) ? R.coarse.n1() : 0; // max summation index
    int j_max = (R.dx2() > 0) ? R.coarse.n2() : 0;

    if (i_max*j_max > 0) // there's something to draw
      {
	// compute and store vertices
	std::vector<std::vector<P> > facets(i_max+1);
	P temp;

	for (int i=0; i <= i_max; ++i)
	  for (int j=0; j <= j_max; ++j)
	    {
	      temp = R.corner1 + (i*R.step1())*E_1 + (j*R.step2())*E_2;
	      facets.at(i).push_back(F(temp.x1(), temp.x2()));
	    }

	std::vector<facet> mesh;

	// store mesh polygons
	for (int i=0; i < i_max; ++i)
	  for (int j=0; j < j_max; ++j)
           mesh.push_back(facet(facets.at(i).at(j),     facets.at(i+1).at(j),
				facets.at(i+1).at(j+1), facets.at(i).at(j+1)));

	sort(mesh.begin(), mesh.end(), by_distance());

	for (unsigned int i=0; i<mesh.size(); ++i)
	  mesh.at(i).draw();
      }
  }


  void surface_rev(double f(double), double g(double), 
		   double min_x, double max_x, int i_max, int j_max)
  {
    double du = (max_x - min_x)/i_max;
    double dv = epix::full_turn()/j_max;
    double u, v=0;

    double tmp_x, tmp_y;

    // compute all vertices
    std::vector<std::vector<P> > facets(i_max+1);

    for (int i=0; i <= i_max; ++i)
      {
	u = min_x + i*du;
	tmp_x = f(u);
	tmp_y = g(u);

	for (int j=0; j <= j_max; ++j)
	  {
	    v = j*dv;
	    // revolve about x-axis
	    facets.at(i).push_back(P(tmp_x, tmp_y*Cos(v), tmp_y*Sin(v)));
	  }
      }

    std::vector<facet> mesh;

    // store mesh polygons
    for (int i=0; i < i_max; ++i)
      for (int j=0; j < j_max; ++j)
	mesh.push_back(facet(facets.at(i).at(j),     facets.at(i+1).at(j),
			     facets.at(i+1).at(j+1), facets.at(i).at(j+1)));

    sort(mesh.begin(), mesh.end(), by_distance());

    for (unsigned int i=0; i<mesh.size(); ++i)
      mesh.at(i).draw();
  }

  void surface_rev(double f(double), double g(double), 
		   const domain& R, const frame& coords)
  {
    double du = R.step1();
    double dv = epix::full_turn()*R.step2();
    double u, v;

    int i_max=R.coarse.n1(), j_max=R.coarse.n2();
    double tmp_x, tmp_y;

    // compute all vertices
    std::vector<std::vector<P> > facets(i_max+1);

    for (int i=0; i <= i_max; ++i)
      {
	u = R.corner1.x1() + i*du;
	tmp_x = f(u);
	tmp_y = g(u);

	for (int j=0; j <= j_max; ++j)
	  {
	    v = R.corner1.x2() + j*dv;
	    // revolve about sea axis
	    facets.at(i).push_back(tmp_x*coords.sea() +
				   tmp_y*Cos(v)*coords.sky() +
				   tmp_y*Sin(v)*coords.eye());
	  }
      }

    std::vector<facet> mesh;

    // store mesh polygons
    for (int i=0; i < i_max; ++i)
      for (int j=0; j < j_max; ++j)
	mesh.push_back(facet(facets.at(i).at(j),     facets.at(i+1).at(j),
			     facets.at(i+1).at(j+1), facets.at(i).at(j+1)));

    sort(mesh.begin(), mesh.end(), by_distance());

    for (unsigned int i=0; i<mesh.size(); ++i)
      mesh.at(i).draw();
  }
} // end of namespace

